"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CellType = exports.Cell = void 0;
const mask_1 = require("./mask");
const helpers_1 = require("../utils/helpers");
const numbers_1 = require("../utils/numbers");
const bits_1 = require("../utils/bits");
const hash_1 = require("../utils/hash");
const HASH_BITS = 256;
const DEPTH_BITS = 16;
var CellType;
(function (CellType) {
    CellType[CellType["Ordinary"] = -1] = "Ordinary";
    CellType[CellType["PrunedBranch"] = 1] = "PrunedBranch";
    CellType[CellType["LibraryReference"] = 2] = "LibraryReference";
    CellType[CellType["MerkleProof"] = 3] = "MerkleProof";
    CellType[CellType["MerkleUpdate"] = 4] = "MerkleUpdate";
})(CellType || (CellType = {}));
exports.CellType = CellType;
class Cell {
    constructor(options) {
        this.hashes = [];
        this.depths = [];
        const { bits = [], refs = [], type = CellType.Ordinary } = options || {};
        this._bits = bits;
        this._refs = refs;
        this._type = type;
        this.initialize();
    }
    get bits() {
        return Array.from(this._bits);
    }
    get refs() {
        return Array.from(this._refs);
    }
    get type() {
        return this._type;
    }
    get exotic() {
        return this._type !== CellType.Ordinary;
    }
    getRefsDescriptor(mask) {
        const value = this._refs.length + (Number(this.exotic) * 8) + ((mask ? mask.value : this.mask.value) * 32);
        const descriptor = Uint8Array.from([value]);
        return (0, helpers_1.bytesToBits)(descriptor);
    }
    getBitsDescriptor() {
        const value = Math.ceil(this._bits.length / 8) + Math.floor(this._bits.length / 8);
        const descriptor = Uint8Array.from([value]);
        return (0, helpers_1.bytesToBits)(descriptor);
    }
    getDepthDescriptor(depth) {
        const descriptor = Uint8Array.from([Math.floor(depth / 256), depth % 256]);
        return (0, helpers_1.bytesToBits)(descriptor);
    }
    getAugmentedBits() {
        return (0, bits_1.augment)(this._bits);
    }
    print(indent = 0) {
        const bits = Array.from(this._bits);
        const areDivisible = bits.length % 4 === 0;
        const augmented = !areDivisible ? (0, bits_1.augment)(bits, 4) : bits;
        const fiftHex = `${(0, helpers_1.bitsToHex)(augmented).toUpperCase()}${!areDivisible ? '_' : ''}`;
        const output = [`${' '.repeat(indent)}x{${fiftHex}}\n`];
        this._refs.forEach(ref => output.push(ref.print(indent + 1)));
        return output.join('');
    }
    hash(level = 3) {
        if (this.type !== CellType.PrunedBranch) {
            return this.hashes[this.mask.apply(level).hashIndex];
        }
        const { hashIndex } = this.mask.apply(level);
        const { hashIndex: thisHashIndex } = this.mask;
        const skip = 16 + (hashIndex * HASH_BITS);
        return hashIndex !== thisHashIndex
            ? (0, helpers_1.bitsToHex)(this._bits.slice(skip, skip + HASH_BITS))
            : this.hashes[0];
    }
    depth(level = 3) {
        if (this.type !== CellType.PrunedBranch) {
            return this.depths[this.mask.apply(level).hashIndex];
        }
        const { hashIndex } = this.mask.apply(level);
        const { hashIndex: thisHashIndex } = this.mask;
        const skip = 16 + (thisHashIndex * HASH_BITS) + (hashIndex * DEPTH_BITS);
        return hashIndex !== thisHashIndex
            ? (0, numbers_1.bitsToIntUint)(this._bits.slice(skip, skip + DEPTH_BITS), { type: 'uint' })
            : this.depths[0];
    }
    eq(cell) {
        return this.hash() === cell.hash();
    }
    validatePrunedBranch() {
        if (this._refs.length !== 0)
            throw new Error('Pruned Branch cell can\'t has refs');
        if (this._bits.length < 16)
            throw new Error('Pruned Branch cell can\'t has less than 16 bits');
        if (this.mask.level < 1 || this.mask.level > 3) {
            throw new Error('Pruned Branch has an invalid level');
        }
        const { hashCount } = this.mask.apply(this.mask.level - 1);
        const size = 8 + 8 + hashCount * (HASH_BITS + DEPTH_BITS);
        if (this._bits.length !== size) {
            throw new Error('Pruned Branch has an invalid data');
        }
    }
    validateLibraryReference() {
        const size = 8 + HASH_BITS;
        if (this._bits.length !== size || this._refs.length !== 0) {
            throw new Error('Library reference has an invalid data');
        }
    }
    validateMerkleProof() {
        const size = 8 + HASH_BITS + DEPTH_BITS;
        if (this._bits.length !== size) {
            throw new Error('Merkle Proof has an invalid data');
        }
        if (this._refs.length !== 1) {
            throw new Error('Merkle Proof has an invalid refs');
        }
        const data = Array.from(this._bits.slice(8));
        const proof = (0, helpers_1.bitsToHex)(data.splice(0, HASH_BITS));
        const hash = this._refs[0].hash(0);
        const depth = (0, numbers_1.bitsToIntUint)(data.splice(0, DEPTH_BITS).reverse(), { type: 'int' });
        if (proof !== hash) {
            throw new Error('Merkle Proof hash mismatch');
        }
        if (depth !== this._refs[0].depth(0)) {
            throw new Error('Merkle Proof depth mismatch');
        }
    }
    validateMerkleUpdate() {
        const size = 8 + (HASH_BITS * 2) + (DEPTH_BITS * 2);
        if (this._bits.length !== size) {
            throw new Error('Merkle Update has an invalid data');
        }
        if (this._refs.length !== 2) {
            throw new Error('Merkle Update has an invalid refs');
        }
        const data = Array.from(this._bits.slice(8));
        for (let i = 0; i < 2; i++) {
            const proof = (0, helpers_1.bitsToHex)(data.splice(0, HASH_BITS));
            const hash = this._refs[i].hash(0);
            if (proof !== hash) {
                throw new Error(`Merkle Update ref #${i} hash mismatch`);
            }
        }
        for (let i = 0; i < 2; i++) {
            const depth = (0, numbers_1.bitsToIntUint)(data.splice(0, DEPTH_BITS), { type: 'uint' });
            if (depth !== this._refs[i].depth(0)) {
                throw Error(`Merkle Update ref #${i} depth mismatch`);
            }
        }
    }
    initialize() {
        switch (this.type) {
            case CellType.Ordinary:
                this.mask = new mask_1.Mask(this._refs.reduce((acc, el) => acc | el.mask.value, 0));
                break;
            case CellType.PrunedBranch:
                this.mask = new mask_1.Mask((0, numbers_1.bitsToIntUint)(this._bits.slice(0, 8), { type: 'uint' }));
                this.validatePrunedBranch();
                break;
            case CellType.LibraryReference:
                this.mask = new mask_1.Mask(0);
                this.validateLibraryReference();
                break;
            case CellType.MerkleProof:
                this.mask = new mask_1.Mask(this._refs[0].mask.value >> 1);
                this.validateMerkleProof();
                break;
            case CellType.MerkleUpdate:
                this.mask = new mask_1.Mask((this._refs[0].mask.value | this._refs[1].mask.value) >> 1);
                this.validateMerkleUpdate();
                break;
            default:
                throw new Error('Unknown exotic cell type');
        }
        const isMerkle = [CellType.MerkleProof, CellType.MerkleUpdate].includes(this.type);
        const isPrunedBranch = this.type === CellType.PrunedBranch;
        const hashCount = isPrunedBranch ? 1 : this.mask.hashCount;
        const hashIndexOffset = this.mask.hashCount - hashCount;
        for (let levelIndex = 0, hashIndex = 0; levelIndex <= this.mask.level; levelIndex++) {
            if (!this.mask.isSignificant(levelIndex))
                continue;
            if (hashIndex < hashIndexOffset) {
                hashIndex++;
                continue;
            }
            if ((hashIndex === hashIndexOffset && levelIndex !== 0 && !isPrunedBranch)
                || (hashIndex !== hashIndexOffset && levelIndex === 0 && isPrunedBranch)) {
                throw new Error('Can\'t deserialize cell');
            }
            const level = levelIndex + Number(isMerkle);
            const bits = hashIndex !== hashIndexOffset
                ? (0, helpers_1.hexToBits)(this.hashes[hashIndex - hashIndexOffset - 1])
                : this.getAugmentedBits();
            let representation = [].concat(this.getRefsDescriptor(this.mask.apply(levelIndex)), this.getBitsDescriptor(), bits);
            const destIndex = hashIndex - hashIndexOffset;
            let depth = 0;
            this.refs.forEach((ref) => {
                const refDepth = ref.depth(level);
                representation = representation.concat(this.getDepthDescriptor(refDepth));
                depth = Math.max(depth, refDepth);
            });
            if (this.refs.length && depth >= 1024) {
                throw new Error('Depth is too big');
            }
            this.refs.forEach((ref) => {
                const refHash = ref.hash(level);
                representation = representation.concat((0, helpers_1.hexToBits)(refHash));
            });
            this.depths[destIndex] = this.refs.length ? depth + 1 : depth;
            this.hashes[destIndex] = (0, hash_1.sha256)((0, helpers_1.bitsToBytes)(representation));
            hashIndex++;
        }
    }
}
exports.Cell = Cell;
//# sourceMappingURL=cell.js.map